@@ -1,7 +1,7 @@
 package Paws::Kinesis::MemoryCaller;
 use 5.008001;
 
-our $VERSION = "0.03";
+our $VERSION = "0.04";
 
 =head1 NAME
 
@@ -76,6 +76,7 @@ with "Paws::Net::CallerRole";
 use namespace::autoclean;
 use Data::UUID;
 use List::AllUtils qw(first_index);
+use MIME::Base64 qw(decode_base64);
 
 use Paws::Kinesis::DescribeStreamOutput;
 use Paws::Kinesis::GetRecordsOutput;
@@ -111,7 +112,7 @@ sub do_call {
         "Paws::Kinesis::GetShardIterator"   => "_get_shard_iterator",
         "Paws::Kinesis::PutRecord"          => "_put_record",
         "Paws::Kinesis::PutRecords"         => "_put_records",
-    }->{$action_class} or die "unknown action ($action_class)";
+    }->{$action_class} or die "($action_class) is not implemented";
 
     $self->$method($action);
 }
@@ -120,9 +121,14 @@ sub _create_stream {
     my $self = shift;
     my ($action) = @_;
 
-    my $shard_count = $action->ShardCount;
+    my $last_shard = $action->ShardCount - 1;
+    $last_shard >= 0
+        or die "ShardCount must be greater than zero to CreateStream";
 
-    my $shard_id__records = { map { $_ => [] } 1..$shard_count };
+    my $shard_id__records = {
+        map { sprintf("shardId-%012d", $_) => [] }
+        0..$last_shard
+    };
 
     $self->store->{$action->StreamName} = $shard_id__records;
 
@@ -136,15 +142,16 @@ sub _get_shard_iterator {
     my $shard_iterator_type = $action->ShardIteratorType;
 
     my $method = {
-        LATEST             => "_get_shard_iterator_latest",
-        TRIM_HORIZON       => "_get_shard_iterator_trim_horizon",
-        AT_SEQUENCE_NUMBER => "_get_shard_iterator_at_sequence_number",
+        LATEST                  => "_get_shard_iterator_latest",
+        TRIM_HORIZON            => "_get_shard_iterator_trim_horizon",
+        AT_SEQUENCE_NUMBER      => "_get_shard_iterator_at_sequence_number",
+        AFTER_SEQUENCE_NUMBER   => "_get_shard_iterator_after_sequence_number",
     }->{$shard_iterator_type}
-        or die "unknown shard_iterator_type ($shard_iterator_type)";
+        or die "ShardIteratorType($shard_iterator_type) is invalid or not implemented";
 
     my $shard_iterator = $self->$method(
-        stream_name => $action->StreamName,
-        shard_id => $action->ShardId,
+        stream_name     => $action->StreamName,
+        shard_id        => $action->ShardId,
         sequence_number => $action->StartingSequenceNumber,
     );
 
@@ -153,20 +160,46 @@ sub _get_shard_iterator {
     );
 }
 
+sub _get_shard_iterator_after_sequence_number {
+    my $self = shift;
+    my %args = @_;
+
+    my $index = $self->_get_index_by_sequence_number(%args);
+
+    return $self->_create_shard_iterator(
+        $args{stream_name},
+        $args{shard_id},
+        $index + 1,
+    );
+}
+
 sub _get_shard_iterator_at_sequence_number {
     my $self = shift;
     my %args = @_;
 
+    my $index = $self->_get_index_by_sequence_number(%args);
+
+    return $self->_create_shard_iterator(
+        $args{stream_name},
+        $args{shard_id},
+        $index,
+    );
+}
+
+sub _get_index_by_sequence_number {
+    my $self = shift;
+    my %args = @_;
+
     my $stream_name = $args{stream_name};
     my $shard_id = $args{shard_id};
-    my $sequence_number = $args{sequence_number};
+    my $sequence_number = $args{sequence_number}
+        or die "StartingSequenceNumber is required";
 
-    my $records = $self->store->{$stream_name}->{$shard_id};
-    my $index = first_index {
+    my $records = $self->_get_records_from_store($stream_name, $shard_id);
+
+    return first_index {
         $_->SequenceNumber eq $sequence_number
     } @$records;
-
-    return $self->_create_shard_iterator($stream_name, $shard_id, $index);
 }
 
 sub _get_shard_iterator_latest {
@@ -176,12 +209,47 @@ sub _get_shard_iterator_latest {
     my $stream_name = $args{stream_name};
     my $shard_id = $args{shard_id};
 
-    my $records = $self->store->{$stream_name}->{$shard_id};
+    my $records = $self->_get_records_from_store($stream_name, $shard_id);
+
     my $index = @$records ? scalar @$records : 0;
 
     return $self->_create_shard_iterator($stream_name, $shard_id, $index);
 }
 
+sub _get_shard_id__records {
+    my $self = shift;
+    my ($stream_name) = @_;
+
+    my $shard_id__records = $self->store->{$stream_name}
+        or die "StreamName($stream_name) does not exist";
+
+    return $shard_id__records;
+}
+
+sub _get_records_from_store {
+    my $self = shift;
+    my ($stream_name, $shard_id) = @_;
+
+    my $shard_id__records = $self->_get_shard_id__records($stream_name);
+
+    my $records = $shard_id__records->{$shard_id}
+        or die sprintf(
+            "ShardId(%s) does not exist. ShardIds are (%s)",
+            $shard_id,
+            join(", ", sort { $a cmp $b } keys %$shard_id__records),
+        );
+
+    return $records;
+}
+
+sub _push_record_to_store {
+    my $self = shift;
+    my ($stream_name, $shard_id, $record) = @_;
+
+    my $records = $self->_get_records_from_store($stream_name, $shard_id);
+    push @$records, $record;
+}
+
 sub _get_shard_iterator_trim_horizon {
     my $self = shift;
     my %args = @_;
@@ -196,7 +264,7 @@ sub _create_shard_iterator {
     my $self = shift;
     my ($stream_name, $shard_id, $index) = @_;
 
-    my $shard_iterator = Data::UUID->new->create_str();
+    my $shard_iterator = Data::UUID->new->create_b64();
 
     $self->shard_iterator__address->{$shard_iterator} = {
         stream_name => $stream_name,
@@ -253,13 +321,16 @@ sub _get_records {
 
     my $address =
         $self->shard_iterator__address->{$shard_iterator}
-            or die "shard_iterator ($shard_iterator) not found";
+            or die "ShardIterator($shard_iterator) does not exist";
 
     my $stream_name = $address->{stream_name};
     my $shard_id = $address->{shard_id};
     my $index = $address->{index};
 
-    my @records = @{$self->store->{$stream_name}->{$shard_id}};
+    my $records = $self->_get_records_from_store($stream_name, $shard_id);
+
+    # make a copy for splice...
+    my @records = @$records;
 
     return Paws::Kinesis::GetRecordsOutput->new(
         Records => [
@@ -278,20 +349,24 @@ sub _put_record {
     my $self = shift;
     my ($action) = @_;
 
-    my $shard_id =
-        $self->_get_shard_id_from_partition_key($action);
+    my $stream_name = $action->StreamName;
+    my $data = $action->Data;
 
-    my $records = $self->store->{$action->StreamName}->{$shard_id};
+    decode_base64($data) && length($data) % 4 == 0
+        or die "Data($data) is not valid Base64";
+
+    my $shard_id = $self->_get_shard_id_from_partition_key($action);
+    my $records = $self->_get_records_from_store($stream_name, $shard_id);
 
     my $sequence_number = scalar(@$records + 1);
 
-    my $new_record = Paws::Kinesis::Record->new(
-        Data => $action->Data,
-        PartitionKey => $action->PartitionKey,
-        SequenceNumber => $sequence_number,
+    my $record = Paws::Kinesis::Record->new(
+        Data            => $data,
+        PartitionKey    => $action->PartitionKey,
+        SequenceNumber  => $sequence_number,
     );
 
-    push @$records, $new_record;
+    $self->_push_record_to_store($stream_name, $shard_id, $record);
 
     return Paws::Kinesis::PutRecordOutput->new(
         ShardId => $shard_id,
@@ -352,10 +427,9 @@ sub _get_shard_ids_from_stream_name {
     my $self = shift;
     my ($stream_name) = @_;
 
-    my $shard_id__records = $self->store->{$stream_name}
-        or die "stream ($stream_name) does not exist";
+    my $shard_id__records = $self->_get_shard_id__records($stream_name);
 
-    return [ sort { $a <=> $b } keys %$shard_id__records ],
+    return [ sort { $a cmp $b } keys %$shard_id__records ],
 }
 
 __PACKAGE__->meta->make_immutable;
